/* * Device driver for the sports arbitrage calculator
 *
 * A Platform device implemented using the misc subsystem
 *
 * Jonathan Nalikka
 * Columbia University
 *
 * References:
 * Linux source: Documentation/driver-model/platform.txt
 *               drivers/misc/arm-charlcd.c
 * http://www.linuxforu.com/tag/linux-device-drivers/
 * http://free-electrons.com/docs/
 *
 * "make" to build
 * insmod calc_arb_ball.ko
 *
 * Check code style with
 * checkpatch.pl --file --no-tree calc_arb.c
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/platform_device.h>
#include <linux/miscdevice.h>
#include <linux/slab.h>
#include <linux/io.h>
#include <linux/of.h>
#include <linux/of_address.h>
#include <linux/fs.h>
#include <linux/uaccess.h>
#include <linux/slab.h>  
#include <linux/printk.h>

#include "../include/arb_ioctl.h"
#include "../include/arb_buf.h"
#include "../include/avalon_interface.h"

#define DRIVER_NAME "calc_arb"

/*
 * macros to get bytes 0->3 of struct arb 
 */
 
#define BYTE_X(word, x) ((char)(((word) >> (8 * x)) & 0xff))
#define EVENT_STRUCT_BYTE_X(event, x) (BYTE_X(*(uint32_t *)(event), x))

/*
 * Information about our device
 */
struct calc_arb_dev {
	struct resource res; /* Resource: our registers */
	void __iomem *virtbase; /* Where registers can be accessed in memory */
} dev;


void print_result(arb_result_t *result)
{
    pr_info("Game ID: %d, Bookie A: %d, Bookie B: %d, Arbitrage Probability: %d\n", result->game_id, result->bookie_id_a, result->bookie_id_b, result->arb_prob);
}

void print_arb_event(arb_event_t event)
{
	pr_info("Outcome: %c, Game ID: %d, Bookie ID: %d, Odds: %x, Unused: %d\n", 
				OC_IS_A(&event) ? 'a' : 'b', event.game_id, event.bookie_id, event.odds, event.unused);
}

/*
 * Write segments of a single digit
 * Assumes digit is in range and the device information has been set up
 */
static void write_events(struct event_buf *buf)
{
    int i;

	// (1) send reset signal
	iowrite32(0, ((uint32_t *)dev.virtbase) + ARB_RESET_ADDR);

	// (2) write events
    for(i = 0; i < buf->len; i++) {
        iowrite32(*(uint32_t *)((buf->events_vec) + i), 
				((uint32_t *)dev.virtbase) + ARB_EVENT_WRITE_ADDR);
    }

	// (3) raise start
    iowrite32(0, ((uint32_t *)dev.virtbase) + ARB_START_ADDR);
}

static struct result_buf *read_result(void)
{
	arb_read_regs_t read_regs;
	int i;
	struct result_buf *results_buf;
	uint32_t readdata;

	// (1) poll for done
	i = 1000;
	while(i--) {
		readdata = ioread32(((uint32_t *)dev.virtbase) + ARB_READ_REGS_ADDR);
		read_regs = *((arb_read_regs_t *) &readdata);

		if (read_regs.done)
			break;
	}

    results_buf = kmalloc(sizeof(int) + read_regs.result_count * sizeof(arb_result_t), GFP_KERNEL);
	results_buf->len = read_regs.result_count;

	// (2) read results structs
	for (i=0; i < results_buf->len; i++) {
		uint32_t readdata = ioread32(((uint32_t *)dev.virtbase) + ARB_RESULT_READ_ADDR(i));
		results_buf->arbs_vec[i] = *((arb_result_t *) &readdata);
	}

	return results_buf;
}

/*
 * Handle ioctl() calls from userspace:
 * Read or write the segments on single digits.
 * Note extensive error checking of arguments
 */
static long calc_arb_ioctl(struct file *f, unsigned int cmd, unsigned long arg)
{
    struct event_buf *buf;
	struct result_buf *res_buf;
    int len;
    size_t size;

	switch (cmd) {
	    case CALC_ARB_WRITE_EVENTS:
		    if (copy_from_user(&(len), (int *) arg,
				       sizeof(int)))
			    return -EACCES;

			size = sizeof(int) + len * sizeof(arb_event_t);
			buf = kmalloc(size, GFP_KERNEL);

			if (copy_from_user(buf, (struct event_buf *) arg,
				       size))
			    return -EACCES;

		    write_events(buf);
		    kfree(buf);
		    break;

	    case CALC_ARB_READ_RESULTS:
	      	res_buf = read_result();
	      	size = sizeof(int) + res_buf->len * sizeof(arb_result_t);
		    if (copy_to_user((struct result_buf *) arg, res_buf,
				     size))
			    return -EACCES;
			kfree(res_buf);
		    break;
	    default:
		    return -EINVAL;
	}

	return 0;
}

/* The operations our device knows how to do */
static const struct file_operations calc_arb_fops = {
	.owner		= THIS_MODULE,
	.unlocked_ioctl = calc_arb_ioctl,
};

/* Information about our device for the "misc" framework -- like a char dev */
static struct miscdevice calc_arb_misc_device = {
	.minor		= MISC_DYNAMIC_MINOR,
	.name		= DRIVER_NAME,
	.fops		= &calc_arb_fops,
};

/*
 * Initialization code: get resources (registers) and display
 * a welcome message
 */
static int __init calc_arb_probe(struct platform_device *pdev)
{
	int ret;

	/* Register ourselves as a misc device: creates /dev/calc_arb */
	ret = misc_register(&calc_arb_misc_device);

	/* Get the address of our registers from the device tree */
	ret = of_address_to_resource(pdev->dev.of_node, 0, &dev.res);
	if (ret) {
		ret = -ENOENT;
		goto out_deregister;
	}

	/* Make sure we can use these registers */
	if (request_mem_region(dev.res.start, resource_size(&dev.res),
			       DRIVER_NAME) == NULL) {
		ret = -EBUSY;
		goto out_deregister;
	}

	/* Arrange access to our registers */
	dev.virtbase = of_iomap(pdev->dev.of_node, 0);
	if (dev.virtbase == NULL) {
		ret = -ENOMEM;
		goto out_release_mem_region;
	}
	//initial write
	//iowrite32(0xffffffff, dev.virtbase);
	return 0;

out_release_mem_region:
	release_mem_region(dev.res.start, resource_size(&dev.res));
out_deregister:
	misc_deregister(&calc_arb_misc_device);
	return ret;
}

/* Clean-up code: release resources */
static int calc_arb_remove(struct platform_device *pdev)
{
	iounmap(dev.virtbase);
	release_mem_region(dev.res.start, resource_size(&dev.res));
	misc_deregister(&calc_arb_misc_device);
	return 0;
}

/* Which "compatible" string(s) to search for in the Device Tree */
#ifdef CONFIG_OF
static const struct of_device_id calc_arb_of_match[] = {
	{ .compatible = "bettor,calc_arb-3.0" },
	{},
};
MODULE_DEVICE_TABLE(of, calc_arb_of_match);
#endif

/* Information for registering ourselves as a "platform" driver */
static struct platform_driver calc_arb_driver = {
	.driver	= {
		.name	= DRIVER_NAME,
		.owner	= THIS_MODULE,
		.of_match_table = of_match_ptr(calc_arb_of_match),
	},
	.remove	= __exit_p(calc_arb_remove),
};

/* Called when the module is loaded: set things up */
static int __init calc_arb_init(void)
{
	pr_info(DRIVER_NAME ": init\n");
	return platform_driver_probe(&calc_arb_driver, calc_arb_probe);
}

/* Calball when the module is unloaded: release resources */
static void __exit calc_arb_exit(void)
{
	platform_driver_unregister(&calc_arb_driver);
	pr_info(DRIVER_NAME ": exit\n");
}

module_init(calc_arb_init);
module_exit(calc_arb_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Jonathan Nalikka, Columbia University");
MODULE_DESCRIPTION("sports arbitrage calculator driver");
