#include <iostream>
#include "Vtwoport_memory.h"
#include <verilated.h>
#include <verilated_vcd_c.h>

#include "../include/arb.h"
#include "../include/arb_buf.h"
#include "../include/arb_parsing.h"

#define MAX_ENTRIES 256

static void update_clk(Vtwoport_memory& mem_sim, VerilatedVcdC& tfp,
	int time)
{
	mem_sim.clk = (time % 20) >= 10;
	mem_sim.eval();
	tfp.dump( time );
}

static void clear_registers_for_port(Vtwoport_memory& mem_sim,
	VerilatedVcdC& tfp, int port=1)
{
	if (port == 1) {
		mem_sim.write1		= 0;
		mem_sim.writedata1	= 0;
	} else {
		mem_sim.write2		= 0;
		mem_sim.writedata2	= 0;
	}
}

static void mem_write(Vtwoport_memory& mem_sim, VerilatedVcdC& tfp,
	uint32_t data, uint16_t addr, int port=1)
{
	if (port == 1) {
		mem_sim.a1			= addr;
		mem_sim.write1		= 1;
		mem_sim.writedata1	= data;
	} else {
		mem_sim.a2			= addr;
		mem_sim.write2		= 1;
		mem_sim.writedata2	= data;
	}
}

static void mem_request_read(Vtwoport_memory& mem_sim, VerilatedVcdC& tfp,
	uint16_t addr, int port=1)
{
	if (port == 1) {
		mem_sim.a1			= addr;
		mem_sim.write1		= 0;
		mem_sim.writedata1	= 0;
	} else {
		mem_sim.a2			= addr;
		mem_sim.write2		= 0;
		mem_sim.writedata2	= 0;
	}
}

static uint32_t mem_finish_read(Vtwoport_memory& mem_sim, VerilatedVcdC& tfp,
	int port=1)
{
	uint32_t res;
	if (port == 1) {
		res = mem_sim.readdata1;
	} else {
		res = mem_sim.readdata2;
	}
	clear_registers_for_port(mem_sim, tfp, port);
	return res;
}

void run_dummy_reads_and_writes(Vtwoport_memory& mem_sim, VerilatedVcdC& tfp)
{
	int time = 0;
	uint32_t res1 = 0, res2 = 0;

	clear_registers_for_port(mem_sim, tfp, 1);
	clear_registers_for_port(mem_sim, tfp, 2);

	for ( ; time < 500; time += 10) {
		// simultaneous writes
		if (time == 120) {
			mem_write(mem_sim, tfp, 0xffff, 1, 1);
			mem_write(mem_sim, tfp, 0xf0f0, 0, 2);
		}
		if (time == 140) {
			clear_registers_for_port(mem_sim, tfp, 1);
			clear_registers_for_port(mem_sim, tfp, 2);
		}

		// simultaneous reads
		if (time == 220) {
			mem_request_read(mem_sim, tfp, 0, 1);
			mem_request_read(mem_sim, tfp, 1, 2);
		}
		if (time == 240) {
			res1 = mem_finish_read(mem_sim, tfp, 1);
			res2 = mem_finish_read(mem_sim, tfp, 2);
		}

		update_clk(mem_sim, tfp, time);
	}
	update_clk(mem_sim, tfp, time);
}

void simulate_actual_reads_and_writes(Vtwoport_memory& mem_sim, VerilatedVcdC& tfp)
{
	int time = 0;
	uint32_t writedata, readdata = 0;
	int write_index = 0, read_index = 0;
	int write_ts = 40, write_clear_ts = 60;
	int read_ts = 140, read_finish_ts = 160;
	arb_event_t *write_event, *read_event;
	struct event_buf *events = events_from_file("../scrape/nba_game_odds_template.csv");
	uint8_t event_count = std::min(MAX_ENTRIES, events->len);

	for ( ; time < 300000 && read_index < event_count; time += 10) {
		// writing an event
		if (time == write_ts && write_index < event_count) {
			write_event = &events->events_vec[write_index];
			writedata = *((uint32_t *)(write_event));
			mem_write(mem_sim, tfp, writedata, write_index); // default to port 1
			write_ts += 100;
			write_index++;
		}
		if (time == write_clear_ts) {
			clear_registers_for_port(mem_sim, tfp);
			write_clear_ts += 100;
		}
		
		// read from port 2 while we keep writing to port 1
		if (time == read_ts && read_index < event_count) {
			mem_request_read(mem_sim, tfp, read_index, 2);
			read_ts += 100;
		}

		// check that what we read matches what we wrote in
		if (time == read_finish_ts) {
			readdata = mem_finish_read(mem_sim, tfp, 2);
			read_event = &events->events_vec[read_index];
			if(memcmp(&readdata, read_event, sizeof(readdata))) {
				std::cerr << "FAILURE\n";
				std::cerr << "\tEvent read at " << read_index << " doesn't match ";
				std::cerr << "event written from " << read_index << "\n";
				std::cerr << "\treaddata: " << readdata << "\n";
				std::cerr << "\toriginal: " << *((uint32_t *)read_event) << "\n";
				exit(EXIT_FAILURE);
			}
			read_index++;
			read_finish_ts += 100;
		}
		update_clk(mem_sim, tfp, time);
	}
	update_clk(mem_sim, tfp, time);

	free(events);
	std::cout << "SUCCESS!\n";
}

int main(int argc, const char **argv, const char **env) {
	Verilated::commandArgs(argc, argv);
	Vtwoport_memory mem_sim;
	Verilated::traceEverOn(true);
	VerilatedVcdC tfp;
	mem_sim.trace(&tfp, 99);
	tfp.open("twoport_memory.vcd");

	// run_dummy_reads_and_writes(mem_sim, tfp);
	simulate_actual_reads_and_writes(mem_sim, tfp);

	tfp.close();
	mem_sim.final();

	return 0;
}

